import subprocess
import os
import gdown
from IPython import display
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from time import time
from sklearn.preprocessing import MinMaxScaler
from keras.preprocessing.sequence import TimeseriesGenerator
import numpy as np

def показать_примеры(база, старт, финиш):
  #Загружаем датафрейм с сессионного хранилища
 
  plt.figure(figsize=(12, 6)) # Создаем полотно для визуализации
  #Переводим тип данных со столбика с ценами из текста (string) на числа с плавающей запятой (float)
  plt.plot(база['Трафик'][старт:финиш], label = 'Трафик')
  plt.legend()
  plt.show()

def создать_выборки_трафик(база, xLen):
  batch_size=20
  curTime = time()
  #Загружаем данные в переменную dataframe
  data = база.iloc[:,1]
  data = np.array(data) #Превращаем в numpy массив
  for i in range(len(data)):
    data[i] = float(data[i])
  
  #Разделяем на проверочную/обучаюшую выборки с соотношением 20%/80%
  #Формируем параметры загрузки данных
  valLen = 300  # Используем 300 записей для проверки

  trainLen = data.shape[0]-valLen # Размер тренировочной выборки

  # Делим данные на тренировочную и тестовую выборки 
  Train,Test = np.reshape(data[:trainLen],(-1,1)), np.reshape(data[trainLen+xLen+2:],(-1,1))

  # Масштабируем данные (отдельно для X и Y), чтобы их легче было скормить сетке
  Scaler = MinMaxScaler()
  Scaler.fit(Train)
  Train = Scaler.transform(Train)
  Test = Scaler.transform(Test)


  # Создаем генератор для обучения
  trainDataGen = TimeseriesGenerator(Train, Train,             # в качестве параметров наши выборки
                                length=xLen, sampling_rate=1, # для каждой точки
                                batch_size=20)                # размер batch, который будем скармливать модели

  # Создаем аналогичный генератор для валидации при обучении
  testDataGen = TimeseriesGenerator(Test, Test,
                                length=xLen, sampling_rate=1,
                                batch_size=20)

  # Создадим генератор проверочной выборки, из которой потом вытащим xVal, yVal для проверки
  DataGen = TimeseriesGenerator(Test, Test,
                                length=xLen, sampling_rate=1,
                                batch_size=len(Test)) # Размер batch будет равен длине нашей выборки

  xVal = []
  yVal = []
  for i in DataGen:
    xVal.append(i[0])
    yVal.append(i[1])

  xVal = np.array(xVal)
  yVal = np.array(yVal)
  print('Выборки созданы успешно')
  return trainDataGen, testDataGen, (xVal, yVal, Scaler)

def тест_модели_трафика(нейронка, наборы):

  def getPred(currModel, xVal, yVal, yScaler):
    # Предсказываем ответ сети по проверочной выборке
    # И возвращаем исходны масштаб данных, до нормализации
    predVal = yScaler.inverse_transform(currModel.predict(xVal))
    yValUnscaled = yScaler.inverse_transform(yVal)
    
    return (predVal, yValUnscaled)
  
  # channel - какой канал отрисовываем
  def showPredict1(start, step, channel, predVal, yValUnscaled):
    plt.figure(figsize=(14,7))
    plt.plot(predVal[start:start+step, channel], 
            label='Прогноз')
    plt.plot(yValUnscaled[start:start+step, channel], 
            label='Базовый ряд')
    plt.xlabel('Время')
    plt.ylabel('Значение Close')
    plt.legend()
    plt.show()

  (predVal, yValUnscaled) = getPred(нейронка, наборы[0][0], наборы[1][0], наборы[2]) #Прогнозируем данные

  # Отображаем графики
  showPredict1(0, 400, 0, predVal, yValUnscaled)


print('Идет проверка и установка дополнительных библиотек, необходимых для работы terra_ai')
print('Это может занять несколько минут...')
import subprocess, os, warnings, time
from pandas.core.common import SettingWithCopyWarning
from subprocess import STDOUT, check_call
from IPython import display
import numpy as np
import requests
import librosa #Для параметризации аудио
import ast
import json
from tabulate import tabulate
import getpass
def выполнить_команду(команда='!ls'):
  proc = subprocess.Popen(f'{команда}', shell=True, stdin=None, stdout=open(os.devnull,"wb"), stderr=STDOUT, executable="/bin/bash")
  proc.wait()
  pass
import os
import time
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' 
import tensorflow as tf
tf.get_logger().setLevel("WARNING")
import logging
logging.getLogger("tensorflow").setLevel(logging.WARNING)
#tf.get_logger().setLevel(logging.ERROR)

выполнить_команду('pip -q install pymorphy2') 
выполнить_команду('pip -q install scikit-learn') 
import sklearn
from sklearn.model_selection import train_test_split #Разбиение на обучающую и проверочную выборку
from sklearn.preprocessing import LabelEncoder, StandardScaler #Для нормировки данных

from . import датасет, констр, модель, сегментация, traide, ряды, прачечная, обработка_текста, сегментация_договоров, квартиры, home, датасет_en, модель_en, сегментация_en, traide_en, обработка_текста_en, квартиры_en, home_en


import seaborn as sns
sns.set_style('darkgrid')
warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

display.clear_output(wait=True)
print('Все необходимые библиотеки установлены и готовы к работе')
  
def load_dataset(dataset_name = '', description = False):
  return датасет.загрузить_базу(
    база = dataset_name,
    справка = description
)
#Демонстрация
def load_ga_moduls():
  print('Выполняется загрузка и установка модулей')
  print('Это может занять несколько минут...')
  выполнить_команду('apt install swig cmake libopenmpi-dev zlib1g-dev > /dev/null 2>&1')
  выполнить_команду('pip -q install stable-baselines==2.5.1 box2d box2d-kengz > /dev/nul 2>&1')
  выполнить_команду('pip -q install gym pyvirtualdisplay > /dev/null 2>&1')
  выполнить_команду('pip -q install xvfbwrapper > /dev/null 2>&1')
  выполнить_команду('apt-get update > /dev/null 2>&1')
  выполнить_команду('sudo apt-get install xvfb > /dev/null 2>&1')
  выполнить_команду('apt-get install xdpyinfo > /dev/null 2>&1')
  display.clear_output(wait=True)
  from . import генетика
  print('Все модули установлены и готовы к работе')

def create_display(width, hight):
  from . import генетика
  ширина = width
  высота = hight
  генетика.создать_дисплей(ширина, высота)

def ga_example(task):
  from . import генетика
  задача = task
  генетика.показать_пример(задача)

def fit_ga(population_size, epochs, mutation_rate, survivors):
  from . import генетика
  размер_популяции = population_size
  количество_эпох = epochs
  коэфицент_мутации = mutation_rate
  количество_выживших = survivors
  генетика.обучить_алгоритм(размер_популяции, количество_эпох, коэфицент_мутации, количество_выживших)

def landing_ship():
  from . import генетика
  генетика.посадить_корабль()
  
def конструктор():
  return констр.конструктор()

def mnist_demo():
  return модель_en.демонстрация_МНИСТ()

def show_mask_mnist(number):
  return модель_en.показать_изменение_маски_MNIST(number)

def show_chenging_weights():
  return модель_en.показать_изменение_веса()

def auto_demo():
  return модель_en.демонстрация_АВТО()

def show_mask_auto(num):
  return модель_en.показать_изменение_маски_АВТО(num)

def show_masks():
  import logging
  tf.get_logger().setLevel(logging.ERROR)
  return модель_en.показать_маски()

#LabStory
# Авторизация
def авторизация_LabStory():
  return модель.авторизация_LabStory()  
  
# Добавление набора данных
def добавить_датасет_LabStory(dataset_dict):
  return модель.добавить_датасет_LabStory(dataset_dict)
  
# Получение списка набора данных
def список_датасетов_LabStory():
  return модель.список_датасетов_LabStory()

# Удаление набора данных
def удалить_датасет_LabStory(id):
  return модель.удалить_датасет_LabStory(id)

#Выбор датасета по id
def выбрать_датасет_LabStory(id):
  return модель.выбрать_датасет_LabStory(id)

# Вывод информации о текущем датасете
def текущий_датасет():
  return модель.текущий_датасет()

# Создание задачи    
def добавить_задачу_LabStory(task_dict):
  return модель.добавить_задачу_LabStory(task_dict)

# Получение списка задач 
def список_задач_LabStory():
  return модель.список_задач_LabStory()

# Выбор задачи по id    
def выбрать_задачу_LabStory(id):
  return модель.выбрать_задачу_LabStory(id)

def текущая_задача():
  return модель.текущая_задача()

# Сохранение эксперимента
def сохранить_эксперимент_LabStory(experiment_dict):
  return модель.сохранить_эксперимент_LabStory(experiment_dict)
  
def сохранить_эксперимент_LabStory_terra_ai(experiment_dict):
  return модель.сохранить_эксперимент_LabStory_terra_ai(experiment_dict)

# Получение списка экспериметов
def список_экспериметов_LabStory():
  return модель.список_экспериметов_LabStory()

# Получение эксперимета по id  
def посмотреть_эксперимент_по_id_LabStory(id):
  return модель.посмотреть_эксперимент_по_id_LabStory(id)
  
def все_эксперименты_по_задаче(id):
    return модель.все_эксперименты_по_задаче(id)

def получить_архитектуру(id_exp):
    return модель.получить_архитектуру(id_exp)
    
## Прачечная
def input_data():
  прачечная.ввод_данных()
def make_day_plan(общее_число_ботов, количество_выживших, количество_эпох, коэффициент_мутаций):
  прачечная.рассчитать_план_на_день(общее_число_ботов, количество_выживших, количество_эпох, коэффициент_мутаций)
def show_plan_ga():
  прачечная.план_на_день()
def bags_info(n=None):
  прачечная.информация_о_мешке(n)
  
### ДАТАСЕТ
def examples(**kwargs):
  датасет_en.показать_примеры(**kwargs)

def contract_examples(количество = 1):
    сегментация_договоров.показать_пример(количество)

def time_series_examples(база, старт, финиш):
    return ряды.показать_примеры(база, старт, финиш)

def flat_examples(quantity = 1):
    количество = quantity
    квартиры_en.samples(количество)

def smart_home_examples():
    home_en.показать_примеры_голосовых_команд()

def traide_examples(данные):
    traide_en.show_full_data(данные)

def traide_example_intervals(данные, start, end):
    traide_en.show_data(данные, start, end)

def preprocessing_data(**kwargs):
  return датасет_en.предобработка_данных(**kwargs)

def get_input_size(arr):
  return arr.shape[1:]

def get_output_size(arr):
  return arr.max()+1

def create_array(path, size, коэф_разделения=0.9):
  путь = path
  размер = size
  return датасет_en.создать_выборки(путь, размер, коэф_разделения)

def create_txt_array(MWC, xLen, step, путь_к_базе='content/Болезни/'):
  return обработка_текста.создать_выборки(MWC, xLen, step, путь_к_базе)

def create_flat_array():
  return квартиры_en.создать_выборки()

def create_traide_array(shares,analyzed_days,period):
  акции = shares
  количество_анализируемых_дней = analyzed_days
  период_предсказания = period
  return traide_en.получить_данные(акции,
  количество_анализируемых_дней,
  период_предсказания)

def create_smart_home_array(длина, шаг):
  return home_en.создать_выборки(длина, шаг)

def create_time_series_array(база, период):
  return ряды.создать_выборки_трафик(база, период)

def create_txt_segmentation_array(договоры):
  return сегментация_договоров.создать_выборки_договоров(договоры)
  
def create_chat_bot_array(вопросы, ответы,количество_пар=10000):
  return обработка_текста.создать_выборки_чатбота(вопросы, ответы,количество_пар)
  
def create_segmentation_array(images_airplane, segments_airplane):
   return сегментация_en.create_xy(images_airplane, segments_airplane)

def create_high_resolution_array():
  return повышение_размерности.генератор_данных_DIV2K()

# МОДЕЛЬ
def show_plan(мод):
  return модель_en.схема_модели(мод)

def fit_flat_model(мод, x_train, y_train, x_test=[], y_test=[], batch_size=None, epochs=None, коэф_разделения = 0.2, scaler = None):
  размер_пакета = batch_size
  количество_эпох = epochs
  инструменты = scaler
  return модель_en.обучение_модели_квартиры(мод, x_train, y_train, x_test, y_test, размер_пакета, количество_эпох, коэф_разделения, инструменты)

def fit_model(мод, x_train, y_train, x_test=[], y_test=[], batch_size=None, epochs=None, коэф_разделения = 0.2):
  global history
  размер_пакета = batch_size
  количество_эпох = epochs
  return модель_en.обучение_модели(мод, x_train, y_train, x_test, y_test, размер_пакета, количество_эпох, коэф_разделения)

def fit_time_series_model(мод, ген1, ген2, количество_эпох=None):
  модель_en.обучение_модели_трафик(мод, ген1, ген2, количество_эпох)
  return history

def fit_generator(генератор, обучающая_выборка, проверочная_выборка, количество_шагов, интервал_вывода):
  повышение_размерности.предобучение_генератора(генератор,обучающая_выборка, проверочная_выборка, количество_шагов, интервал_вывода)
  print('Предобучение генератора завершено')
  
def fit_high_resolution_model(нейронка,обучающая_выборка):
  нейронка.train(обучающая_выборка, steps=5000) #Обучаем Srgan полностью

def fit_detection_model(количество_эпох):
  обнаружение.обучение_модели_обнаружения(количество_эпох)

def test_classification(net, test_images, labels, classes, quantity=1):
  нейронка = net
  тестовый_набор = test_images
  правильные_ответы = labels
  классы = classes
  количество = quantity
  return модель_en.тест_модели_классификации(нейронка, тестовый_набор, правильные_ответы, классы, количество)

def test_symptoms(нейронка, xLen, step, симптомы, classes):
    классы = classes
    обработка_текста_en.тест_модели_симп(нейронка, xLen, step, симптомы, классы)

def test_writers(нейронка, xLen, step, текст, классы):
    обработка_текста_en.тест_модели_писатели(нейронка, xLen, step, текст, классы)
    
def test_own_image(net, size, classes):
  нейронка = net
  размер_изображения = size
  классы = classes
  return модель_en.тест_на_своем_изображении(нейронка, размер_изображения, классы)

def test_segmentation(мод, test_images,  **kwargs):
  тестовые_изображения = test_images
  сегментация_en.тест_модели(мод, тестовые_изображения,  **kwargs)

def test_segmentation_contract(мод, теги, договора):
  сегментация_договоров.тест_модели(мод, теги, договора)

def test_traide(нейронка,тестовая_выборка, метки_тестовой_выборки, данные, период_предсказания,количество_анализируемых_дней):
  traide_en.model_test(
    нейронка, 
    тестовая_выборка, метки_тестовой_выборки,
    данные,
    период_предсказания,
    количество_анализируемых_дней)

def recognizing_examples(нейронка_1, период_предсказания,количество_анализируемых_дней):
    traide_en.example_traid(нейронка_1, период_предсказания,количество_анализируемых_дней)

def test_high_resolution(нейронка):
  модель_en.тест_модели_HR(нейронка)

def trading(нейронка_1, тестовая_выборка, данные, type_traide):
  тип = type_traide
  traide_en.traiding(нейронка_1, тестовая_выборка, данные, тип)

def test_chat_bot(нейронка, размер_словаря, энкодер, декодер):
  обработка_текста_en.тест_модели_чат_бот(нейронка, размер_словаря, энкодер, декодер)
  
def test_trained_chat_bot():
  обработка_текста_en.загрузить_предобученную_модель()
  
def test_smart_home(нейронка, limit, length, step):
  порог = limit
  длина = length
  шаг = step
  home_en.тест_модели(нейронка, порог, длина, шаг)

def test_detection(нейронка):
  обнаружение.тест_модели()

def test_time_series(мод, данные):
  ряды.тест_модели_трафика(мод, данные)

def test_flat(нейронка, инструменты, *данные,):
  квартиры_en.тест_модели(нейронка, инструменты, *данные,)

def create_network(layer, input_size, task, параметры_модели=None):
  слои = layer
  входной_размер = input_size
  задача = task
  return модель_en.создать_сеть(слои, входной_размер, параметры_модели, задача)

def create_composite_network(данные, метки, *нейронки):
  return модель_en.создать_составную_сеть(данные, метки, *нейронки)

def create_flat_composite_network(данные, *нейронки):
  return модель_en.создать_составную_сеть_квартиры(данные, *нейронки)

def create_Sequence_to_Sequence(размер_словаря, энкодер, декодер):
  return модель_en.создать_сеть_чат_бот(размер_словаря, энкодер, декодер)

def create_generator(стартовый_блок, основной_блок, финальный_блок):
  мод = модель.создать_генератор_повышения_размерности(стартовый_блок, основной_блок, финальный_блок)
  предобучение_генератора = повышение_размерности.SrganGeneratorTrainer(мод, checkpoint_dir=f'folder')    
  print('Модель генератора создана')    
  return предобучение_генератора

def create_discriminator(блок_дискриминатора,количество_блоков, финальный_блок):
  return модель_en.создать_дискриминатор_повышения_размерности(блок_дискриминатора,количество_блоков, финальный_блок)

def create_PSP(**kwargs):
  return модель_en.создать_PSP(**kwargs)

def create_UNET(**kwargs):
  return модель_en.создать_UNET(**kwargs)

def create_high_resolution(генератор, дискриминатор):
  return повышение_размерности.создать_модель(генератор, дискриминатор)

def load_high_resolution():
  return повышение_размерности.загрузить_веса_готовой_модели()

def load_object_detection_model():
  return обнаружение.cоздать_модель_YOLO()  

def audio_parameterization(comand):
  файл = comand
  home_en.параметризация_аудио(файл)


from . import обнаружение
from IPython import display
import gym # здесь лежат все модели для 
from gym import logger as gymlogger # загрузим библиотеку для изменения параметра вывода ошибок
from gym.wrappers import Monitor # библиотека для обертки видео из хранилища в ячейку
gymlogger.set_level(40) # для правильного отображения ошибок
import numpy as np # библиотека массивов numpy
import random # библиотека для генерации случайных значений
import matplotlib # библиотека для визуализации процессов
import matplotlib.pyplot as plt # библиотека для построения графика
import glob # расширение для использования Unix обозначений при задании пути к файлу
import io # библиотека для работы с потоковыми данными
import base64 # расширение для преобразования в формат base64 (универсальный формат хранения сырых изображений в виде набора электрических сигналов)
from IPython.display import HTML # библиотека для кодирования в код HTML
import time # библиотека для расчета времени обучения
from IPython import display as ipythondisplay # для работы с "сырым" форматом (набор сигналов, а не пиксели)
from pyvirtualdisplay import Display # для создания окна дисплея

def загрузить_модули():  
  print('Выполняется загрузка и установка модулей')
  print('Это может занять несколько минут...')
  обнаружение.выполнить_команду('apt install swig cmake libopenmpi-dev zlib1g-dev > /dev/null 2>&1')
  обнаружение.выполнить_команду('pip -q install stable-baselines==2.5.1 box2d box2d-kengz > /dev/nul 2>&1')
  обнаружение.выполнить_команду('pip -q install gym pyvirtualdisplay > /dev/null 2>&1')
  обнаружение.выполнить_команду('pip -q install xvfbwrapper > /dev/null 2>&1')
  обнаружение.выполнить_команду('apt-get update > /dev/null 2>&1')
  обнаружение.выполнить_команду('sudo apt-get install xvfb > /dev/null 2>&1')
  обнаружение.выполнить_команду('apt-get install xdpyinfo > /dev/null 2>&1')
  display.clear_output(wait=True)
  print('Все модули установлены и готовы к работе')
  
def создать_дисплей(ширина, высота):
  display = Display(visible=0, size=(ширина, высота))
  display.start()
  print('Виртуальный дисплей создан')

def показать_пример(задача):
  def show_video():
    mp4list = glob.glob('video/*.mp4')
    if len(mp4list) > 0:
      mp4 = mp4list[0]
      video = io.open(mp4, 'r+b').read()
      encoded = base64.b64encode(video)
      ipythondisplay.display(HTML(data='''<video alt="test" autoplay 
                  loop controls style="height: 400px;">
                  <source src="data:video/mp4;base64,{0}" type="video/mp4" />
              </video>'''.format(encoded.decode('ascii'))))
    else: 
      print("Could not find video")
    

  def wrap_env(env):
    env = Monitor(env, './video', force=True)
    return env

  env = gym.make('LunarLander-v2')
  
  numBots = 100
  popul = [] # здесь будет лежать популяция

  for i in range(numBots):  
    bot = np.random.random((8,4))
    popul.append(bot)

  env = wrap_env(env) # оборачиваем наше окружения для записи видео
  env.seed(0) # симуляция одного и того же окружения
  observation = env.reset() # обнуляем вектор наблюдения
  i=0
  action = 0
  while True:

      env.render() # Рендер окружения
      # В качестве действия будем подавать значения нашего лучшего бота
      observation,reward,done,_ = env.step(action) # остлеживаем все параметры для подсчета функции значения
      result = np.dot(observation,popul[np.random.randint(0, numBots-1)])
      action = np.argmax(result)
      if done: 
        break;

  
  env.close()
  show_video()

def обучить_алгоритм(размер_популяции, количество_эпох, коэфицент_мутации, количество_выживших):
  
  '''
    Функция получения выжившей популяции
        Входные параметры:
        - popul - наша популяция
        - val - текущие значения
        - nsurv - количество выживших
        - reverse - указываем требуемую операцию поиска результата: максимизация или минимизация
  '''
  def getSurvPopul(
          popul,
          val,
          nsurv,
          reverse
          ):
      newpopul = [] # Двумерный массив для новой популяции

      sval= sorted(val, reverse=reverse) # Сортируем зачения в val в зависимости от параметра reverse
      for i in range(nsurv): # Проходимся по циклу nsurv-раз (в итоге в newpopul запишется nsurv-лучших показателей)
          index = val.index(sval[i]) # Получаем индекс i-того элемента sval в исходном массиве val
          newpopul.append(popul[index]) # В новую папуляцию добавляем элемент из текущей популяции с найденным индексом
      return newpopul, sval # Возвращаем новую популяцию (из nsurv элементов) и сортированный список

  '''
    Функция получения родителей
        Входные параметры:
        - curr_popul - текущая популяция
        - nsurv - количество выживших
'''
  def getParents(
          curr_popul,
          nsurv
          ):   
      indexp1 = random.randint(0, nsurv - 1) # Случайный индекс первого родителя в диапазоне от 0 до nsurv - 1
      indexp2 = random.randint(0, nsurv - 1) # Случайный индекс второго родителя в диапазоне от 0 до nsurv - 1    
      botp1 = curr_popul[indexp1] # Получаем первого бота-родителя по indexp1
      botp2 = curr_popul[indexp2] # Получаем второго бота-родителя по indexp2 
    
      return botp1,botp2 # Возвращаем обоих полученных ботов

  '''
    Функция смешивания (кроссинговера) двух родителей
        Входные параметры:
        - botp1 - первый бот-родитель
        - botp2 - второй бот-родитель
        - j - номер компонента бота
'''

  def crossPointFrom2Parents(
          botp1,
          botp2, 
          j
          ):
      pindex = np.random.random() # Получаем случайное число в диапазоне от 0 до 1
      
      #Если pindex меньше 0.5, то берем значения от первого бота, иначе от второго
      if pindex < 0.5:
          x = botp1[j]
      else:
          x = botp2[j]
      return x # Возвращаем значние бота  

  '''
    Функция расчета вознаграждения за эпизод
        Входные параметры:
        - popul - популяция ботов
'''


  def countValue(
      popul
      ):
    
    action=0  # генерируем первое  действие случайным из пространства действий
                                      # 0 - ничего не делать 
                                      # 1 - запустить двигатель с левой ориентацией
                                      # 2 - запустить двигатель по центру
                                      # 3 - запустить двигатель с правой ориентацией

    reward_list=[] # здесь будет сумма вознаграждений для каждого эпизода

    for bot in popul: # проходимся по каждому боту в популяции
      
      env.reset() # И обновлять окружение
      i=0 
      tmp=0
      done=False
      while done!=True:
        observation,reward,done,_ = env.step(action) # остлеживаем все параметры для подсчета функции значения
        result = np.dot(observation,bot) # матрично перемножаем бота и вектор наблюдения для предсказания следующего движения
        action = np.argmax(result) # максимальный аргумент - наше движение
        tmp+=reward
      reward_list.append(tmp) # Функция, по которой будет вычисляться "успех текущей симуляции"
      
    return reward_list


  env = gym.make('LunarLander-v2')
  total=[] # Для построения графика
  midle=[] # Для построения графика
  worst=[] # Для построения графика
  n = размер_популяции
  nsurv = количество_выживших
  nnew = n-nsurv # количество новых
  epohs = количество_эпох
  mut = коэфицент_мутации
  curr_time = time.time()
  
  global popul
  global newpopul

  numBots = 100
  popul = [] # здесь будет лежать популяция

  for i in range(numBots):  
    bot = np.random.random((8,4))
    popul.append(bot)

  for it in range(epohs): # создали список списков всех значений по эпохам
    val = countValue(popul) # считаем успех каждого из ботов
    newpopul, sval = getSurvPopul(popul, val, nsurv,1) # получили популяцию выживших, нас интересует бот с максимальным успехом, поэтому reverse = 1
    print('Выполняется эпоха №' '{:>2} {:>3}'.format(it,'Время на выполнение:' ), '{:>4}'.format(np.round(time.time() - curr_time, 1)), 'сек,', '{:>30}'.format('Награда для 3-x лучших ботов'), np.array2string(np.round(sval[0:3], 2).astype('int'),  precision=2,suppress_small=True)) # Выводим время на операцию, среднее значение и 3 лучших ботов
    total.append(sval[0]) # заносим самого лучшего бота в список для построения графика эволюции 
    midle.append(sval[len(sval) // 2])
    worst.append(sval[-1])
    curr_time = time.time() # Обновляем текущее время

  # проходимся по новой популяции
    for k in range(nnew):
      
      # вытаскиваем новых родителей
      botp1, botp2 = getParents(newpopul, nsurv) 
      newbot = [] # здесь будет новый бот
      
      for j in range(len(botp1)): # боты-родители одинаковой длины, будем проходиться по каждому элементу родителя
        x = crossPointFrom2Parents(botp1, botp2,j) # скрещиваем
        for t in range(4):
          if random.random()<mut:
            x[t] += random.random()*1e-1
        newbot.append(x) # закидываем элемент в бота
      newpopul.append(newbot) # добавляем бота
      
    popul = newpopul # вывести список на эпоху
    popul = np.array(popul) # для того, чтобы можно было легко вытащить индексы условием, преобразуем в numpy массив

  plt.plot(total, label='Лучший')
  plt.plot(midle, label='Средний')
  plt.plot(worst, label='Худший')
  plt.yticks(np.arange(-2600, 600, 200))
  plt.legend()
  plt.show()
  print('Среднее количество баллов', round(np.mean(total), 2))  

def посадить_корабль():
  def show_video():
    mp4list = glob.glob('video/*.mp4')
    if len(mp4list) > 0:
      mp4 = mp4list[0]
      video = io.open(mp4, 'r+b').read()
      encoded = base64.b64encode(video)
      ipythondisplay.display(HTML(data='''<video alt="test" autoplay 
                  loop controls style="height: 400px;">
                  <source src="data:video/mp4;base64,{0}" type="video/mp4" />
              </video>'''.format(encoded.decode('ascii'))))
    else: 
      print("Could not find video")
      
  env = gym.make('LunarLander-v2')
  def wrap_env(env):
    env = Monitor(env, './video', force=True)
    return env
  
  env = wrap_env(env) # оборачиваем наше окружения для записи видео
  observation = env.reset() # обнуляем вектор наблюдения
  i=0
  action = 0
  reward = 0
  while True:
    
    env.render() # Рендер окружения
    # В качестве действия будем подавать значения нашего лучшего бота
    observation,reward,done,_ = env.step(action) # остлеживаем все параметры для подсчета функции значения
    result = np.dot(observation, popul[0])
    action = np.argmax(result)
    if done:       
      break

  env.close()
  show_video()
  
  if reward == 100:
    print('Поздравляем, вы посадили корабль! ', 'Награда: ', reward, 'баллов')
  elif reward == -100:
    print('Корабль разбился, попробуйте еще раз! ', 'Награда: ', reward, 'баллов')



import os, nltk, pymorphy2, io, json
from tensorflow.keras.preprocessing.text import Tokenizer, text_to_word_sequence
from tensorflow.keras.preprocessing.text import tokenizer_from_json
from tensorflow.keras import utils
from tensorflow.keras.models import Model, load_model
from keras.layers import *
from tensorflow.keras.preprocessing.sequence import pad_sequences
import numpy as np
from IPython import display
nltk.download('stopwords')
import importlib.util, sys, gdown
import requests
import ast
import json

def readText(fileName, encod):
    f = open(fileName, 'r', encoding=encod)
    text = f.read()
    text = text.replace("\n", " ")
    return text
    
def создать_выборки(maxWordsCount, xLen, step, путь_к_базе='content/Болезни/'):
  print('Происходит создание выборки для обучения')
  print('Это может занять несколько минут...')
  path = путь_к_базе
  text = []
  classes = []
  n = 0
  codecs_list = ['UTF-8', 'Windows-1251']

  for filename in sorted(os.listdir(path)): # Проходим по всем файлам в директории договоров
      n +=1
      for codec_s in codecs_list:
        try:
            text.append(readText(path+filename, codec_s)) # Преобразуем файл в одну строку и добавляем в agreements
            classes.append(filename.replace(".txt", ""))
            break
        except UnicodeDecodeError:
            print('Не прочитался файл: ', path+currdir+'/'+filename, codec_s)
        else:
            next 

  stop_words = nltk.corpus.stopwords.words('russian')
  lexeme_list = ['POS', 'animacy', 'aspect', 'case', 'gender', 'involvement', 'mood', 'number', 'person', 'tense', 'transitivity', 'voice']

  words = [] # Здесь будут лежать все списки слов каждого из описаний заболеваний
  tags = []   # Здесь будут лежать все списки списков граммем для каждого слова
  tags_all = [] # Здесь будут лежать все списки граммем всех слов для тренировки токенайзера
  for i in range(len(text)):
    word, tag = text2Words(text[i])
    words.append(word)
    tags.append(tag)
  for k in tags:
    for t in k:
      tags_all.append(t)

  #################
  #Преобразовываем текстовые данные в числовые/векторные для обучения нейросетью
  #################

  # Максимальное количество слов в словаре
  # Токенизатор кераса
  tokenizer = Tokenizer(num_words=maxWordsCount, filters='!"#$%&()*+,-––—./:;<=>?@[\\]^_`{|}~\t\n\xa0', lower=True, split=' ', oov_token='unknown', char_level=False)
  # Скармливаем ему слова
  tokenizer.fit_on_texts(words) 
  items = list(tokenizer.word_index.items())


  tokenizer_json1 = tokenizer.to_json()
  with io.open('tokenizer1.json', 'w', encoding='utf-8') as f:
      f.write(json.dumps(tokenizer_json1, ensure_ascii=False))
  with open('tokenizer1.json') as f:
      data = json.load(f)
      tokenizer = tokenizer_from_json(data)

  items = list(tokenizer.word_index.items()) 
  #Выведем первые 10 слов из словаря

  # Максимальное количество слов в словаре
  maxWordsCount2 = 50 
  # Токенизатор кераса
  tokenizer2 = Tokenizer(num_words=maxWordsCount2, filters='!"#$%&()*+,-––—./:;<=>?@[\\]^_`{|}~\t\n\xa0', lower=True, split=' ', oov_token='unknown', char_level=False)
  # Скармливаем ему слова
  tokenizer2.fit_on_texts(tags_all) 
  items2 = list(tokenizer2.word_index.items()) 

  tokenizer_json2 = tokenizer2.to_json()
  with io.open('tokenizer2.json', 'w', encoding='utf-8') as f:
      f.write(json.dumps(tokenizer_json2, ensure_ascii=False))

  with open('tokenizer2.json') as f:
      data = json.load(f)
      tokenizer2 = tokenizer_from_json(data)

  items2 = list(tokenizer2.word_index.items())

  #Преобразовываем текст в последовательность индексов согласно частотному словарю
  xTrainIndexes = tokenizer.texts_to_sequences(words) #Обучающие тексты в индексы
  #Преобразовываем тэги в последовательность индексов согласно частотному словарю
  xTrainTagsIndexes = []
  for tag in tags:  # так как теги имеют дополнительные вложенные списки поэтому итерируем в нужном для токенайзера формате
    xTrainTagsIndexes.append(tokenizer2.texts_to_sequences(tag))

  nVal = 200   # Количество слов проверочной выборки

  trainWords = []  # Здесь будет лежать слова для обучающей выборки
  valWords = []    # Здесь будет лежать слова для проверочной выборки

  for i in range(len(xTrainIndexes)):
    trainWords.append(xTrainIndexes[i][:-nVal])
    valWords.append(xTrainIndexes[i][-nVal:])

  trainTagsWords = []  # Здесь будет лежать теги для обучающей выборки
  valTagsWords = []    # Здесь будет лежать тэги для проверочной выборки

  for i in range(len(xTrainTagsIndexes)):
    trainTagsWords.append(xTrainTagsIndexes[i][:-nVal])
    valTagsWords.append(xTrainTagsIndexes[i][-nVal:])

  # Создаем "раскусанные" выборки длины xLen, по 4000 экземпляров на каждый класс 
  (xTrain, yTrain) = createSetsMultiClassesBallanced(trainWords, xLen, step, 4000)
  #(xTrain, yTrain) = createSetsMultiClasses(trainWords, xLen, step)

  (xTrainTags, _) = createSetsMultiClassesBallanced(trainTagsWords, xLen, step, 4000)
  # Преобразовываем полученные обучающие выборки из последовательности индексов в матрицы нулей и единиц по принципу Bag of Words
  xTrain01 = tokenizer.sequences_to_matrix(xTrain.tolist(), mode="tfidf")     # Подаем xTrain в виде списка чтобы метод успешно сработал
  #xTrainTags01 = tokenizer2.sequences_to_matrix(xTrainTags)
  # Создаем "раскусанные" выборки длины xLen для проверочной
  (xVal, yVal) = createSetsMultiClasses(valWords, xLen, step)
  #(xVal, yVal) = createSetsMultiClassesBallanced(valWords, xLen, step, 600)

  (xTagsVal, _) = createSetsMultiClasses(valTagsWords, xLen, step)
  #(xTagsVal, _) = createSetsMultiClassesBallanced(valTagsWords, xLen, step, 600)

  # Преобразовываем полученные выборки из последовательности индексов в матрицы нулей и единиц по принципу Bag of Words
  xVal01 = tokenizer.sequences_to_matrix(xVal.tolist(), mode="tfidf")         # Подаем xVal в виде списка чтобы метод успешно сработал
  #xTagsVal01 = tokenizer2.sequences_to_matrix(xTagsVal.tolist())

  xTrainTags = np.reshape(xTrainTags, (xTrainTags.shape[0], -1))
  xTagsVal = np.reshape(xTagsVal, (xTagsVal.shape[0], -1))
  xTrainTags01 = tokenizer2.sequences_to_matrix(xTrainTags.tolist())
  xTagsVal01 = tokenizer2.sequences_to_matrix(xTagsVal.tolist())

  x_train = [xTrain, xTrain01, xTrainTags01]
  y_train = yTrain
  x_val = [xVal, xVal01, xTagsVal01]
  y_val = yVal
  valw = (valWords, valTagsWords, tokenizer, tokenizer2)
  display.clear_output(wait=True)
  print('Формирование выборки завершено')
  return (x_train, y_train), (x_val, y_val)
  
def text2Words(text):
  stop_words = nltk.corpus.stopwords.words('russian')
  lexeme_list = ['POS', 'animacy', 'aspect', 'case', 'gender', 'involvement', 'mood', 'number', 'person', 'tense', 'transitivity', 'voice']

  text = text.replace(".", " ")
  text = text.replace("—", " ")
  text = text.replace(",", " ")
  text = text.replace("!", " ")
  text = text.replace("?", " ")
  text = text.replace("…", " ")
  text = text.replace("-", " ")
  text = text.replace("(", " ")
  text = text.replace(")", " ")
  text = text.replace(";", " ")
  text = text.replace("°c", " ")
  text = text.replace("–", " ")
  text = text.replace("и/ить", " ")
  text = text.lower()
  morph = pymorphy2.MorphAnalyzer()
  
  words = []
  tags = []
  currWord = ""
  for symbol in text[1:]:

    if (symbol != " "):
      currWord += symbol
    else:
      if (currWord != "") & (currWord not in stop_words): # 
        words.append(currWord)
        currWord = ""
      else:
        currWord = ""
  if (currWord != "") & (currWord not in stop_words): #
        words.append(currWord)
  # генератором списков проходимся по каждому сформированному слову и лемматиризируем его с помощью pymorphy до получения граммем
  #words = [morph.parse(word)[0].normal_form for word in words]
  #tags = [morph.parse(word)[0].tag for word in words] # .cyr_repr
  for word in words:
    tag_word = []
    lex_0 = morph.parse(word)[0].tag.POS  # Part of Speech, часть речи
    if lex_0 != None:
        tag_word.append(lex_0)
    else:
        tag_word.append('not')
    lex_1 = morph.parse(word)[0].tag.animacy  # одушевленность
    if lex_1 != None:
        tag_word.append(lex_1)
    else:
        tag_word.append('not')
    lex_2 = morph.parse(word)[0].tag.aspect # вид: совершенный или несовершенный
    if lex_2 != None:
        tag_word.append(lex_2)
    else:
        tag_word.append('not')
    lex_3 = morph.parse(word)[0].tag.case # падеж
    if lex_3 != None:
        tag_word.append(lex_3)  
    else:
        tag_word.append('not')     
    lex_4 = morph.parse(word)[0].tag.gender # род (мужской, женский, средний)
    if lex_4 != None:
        tag_word.append(lex_4)
    else:
        tag_word.append('not') 
    lex_5 = morph.parse(word)[0].tag.involvement  # включенность говорящего в действие
    if lex_5 != None:
        tag_word.append(lex_5)
    else:
        tag_word.append('not')
    lex_6 = morph.parse(word)[0].tag.mood # наклонение (повелительное, изъявительное)
    if lex_6 != None:
        tag_word.append(lex_6)
    else:
        tag_word.append('not')
    lex_7 = morph.parse(word)[0].tag.number # число (единственное, множественное)
    if lex_7 != None:
        tag_word.append(lex_7)
    else:
        tag_word.append('not')
    lex_8 = morph.parse(word)[0].tag.person # лицо (1, 2, 3)
    if lex_8 != None:
        tag_word.append(lex_8)
    else:
        tag_word.append('not')
    lex_9 = morph.parse(word)[0].tag.tense  # время (настоящее, прошедшее, будущее)
    if lex_9 != None:
        tag_word.append(lex_9)
    else:
        tag_word.append('not')
    lex_10 = morph.parse(word)[0].tag.transitivity  # переходность (переходный, непереходный)
    if lex_10 != None:
        tag_word.append(lex_10)
    else:
        tag_word.append('not')
    lex_11 = morph.parse(word)[0].tag.voice # залог (действительный, страдательный)
    if lex_11 != None:
        tag_word.append(lex_11)
    else:
        tag_word.append('not')
    tags.append(tag_word)
    # генератором списков проходимся по каждому сформированному слову и лемматиризируем его с помощью pymorphy
  words = [morph.parse(word)[0].normal_form for word in words]

  return words, tags
  
def createSetsMultiClassesBallanced(wordIndexes, xLen, step, classSize):

  #Для каждого из 6 классов
  #Создаём обучающую выборку из индексов
  nClasses = len(wordIndexes)
  classesXTrain = []
  for wI in wordIndexes:
    classesXTrain.append(getSetFromIndexes(wI, xLen, step))

  #Формируем один общий xTrain
  xTrain = []
  yTrain = []
  
  for t in range(nClasses):
    xT = classesXTrain[t]
    for i in range(classSize):
      xTrain.append(xT[i % len(xT)])
    
    #Формируем yTrain по номеру класса
    currY = utils.to_categorical(t, nClasses)
    for i in range(classSize):
      yTrain.append(currY)

  xTrain = np.array(xTrain)
  yTrain = np.array(yTrain)
  
  return (xTrain, yTrain)
  
def getSetFromIndexes(wordIndexes, xLen, step):
  xTrain = []
  wordsLen = len(wordIndexes)
  index = 0
  while (index + xLen <= wordsLen):
    xTrain.append(wordIndexes[index:index+xLen])
    index += step
    
  return xTrain
  
def createSetsMultiClasses(wordIndexes, xLen, step):

  #Создаём обучающую выборку из индексов
  nClasses = len(wordIndexes)
  classesXTrain = []
  for wI in wordIndexes:
    classesXTrain.append(getSetFromIndexes(wI, xLen, step))

  #Формируем один общий xTrain
  xTrain = []
  yTrain = []
  
  for t in range(nClasses):
    xT = classesXTrain[t]
    for i in range(len(xT)):
      xTrain.append(xT[i])
    
    #Формируем yTrain по номеру класса
    currY = utils.to_categorical(t, nClasses)
    for i in range(len(xT)):
      yTrain.append(currY)

  xTrain = np.array(xTrain)
  yTrain = np.array(yTrain)

  
  return (xTrain, yTrain)

def тест_модели_симп(нейронка, xLen, step, симптомы, классы):
  gastro_10_predict(нейронка, xLen, step, симптомы, классы)
  
def gastro_10_predict(нейронка, xLen, step, text_input = None, классы = None):
    if text_input is None:
        diagnos =  "Текстовое поле пустое"
    
    else:
        classes = классы
        nClasses = len(классы)
        stop_words = nltk.corpus.stopwords.words('russian')
        lexeme_list = ['POS', 'animacy', 'aspect', 'case', 'gender', 'involvement', 'mood', 'number', 'person', 'tense', 'transitivity', 'voice']

        # Функция очистки текста и превращение в набор слов
        def text2Words(text):
          text = text.replace(".", " ")
          text = text.replace("—", " ")
          text = text.replace(",", " ")
          text = text.replace("!", " ")
          text = text.replace("?", " ")
          text = text.replace("…", " ")
          text = text.replace("-", " ")
          text = text.replace("(", " ")
          text = text.replace(")", " ")
          text = text.replace(";", " ")
          text = text.replace("°c", " ")
          text = text.replace("–", " ")
          text = text.replace("и/ить", " ")
          text = text.lower()
          morph = pymorphy2.MorphAnalyzer()
          
          words = []
          tags = []
          currWord = ""
          for symbol in text[1:]:

            if (symbol != " "):
              currWord += symbol
            else:
              if (currWord != "") & (currWord not in stop_words): # 
                words.append(currWord)
                currWord = ""
              else:
                currWord = ""
          if (currWord != "") & (currWord not in stop_words): #
                words.append(currWord)
          # генератором списков проходимся по каждому сформированному слову и лемматиризируем его с помощью pymorphy до получения граммем
          for word in words:
            tag_word = []
            lex_0 = morph.parse(word)[0].tag.POS  # Part of Speech, часть речи
            if lex_0 != None:
                tag_word.append(lex_0)
            else:
                tag_word.append('not')
            lex_1 = morph.parse(word)[0].tag.animacy  # одушевленность
            if lex_1 != None:
                tag_word.append(lex_1)
            else:
                tag_word.append('not')
            lex_2 = morph.parse(word)[0].tag.aspect # вид: совершенный или несовершенный
            if lex_2 != None:
                tag_word.append(lex_2)
            else:
                tag_word.append('not')
            lex_3 = morph.parse(word)[0].tag.case # падеж
            if lex_3 != None:
                tag_word.append(lex_3)  
            else:
                tag_word.append('not')     
            lex_4 = morph.parse(word)[0].tag.gender # род (мужской, женский, средний)
            if lex_4 != None:
                tag_word.append(lex_4)
            else:
                tag_word.append('not') 
            lex_5 = morph.parse(word)[0].tag.involvement  # включенность говорящего в действие
            if lex_5 != None:
                tag_word.append(lex_5)
            else:
                tag_word.append('not')
            lex_6 = morph.parse(word)[0].tag.mood # наклонение (повелительное, изъявительное)
            if lex_6 != None:
                tag_word.append(lex_6)
            else:
                tag_word.append('not')
            lex_7 = morph.parse(word)[0].tag.number # число (единственное, множественное)
            if lex_7 != None:
                tag_word.append(lex_7)
            else:
                tag_word.append('not')
            lex_8 = morph.parse(word)[0].tag.person # лицо (1, 2, 3)
            if lex_8 != None:
                tag_word.append(lex_8)
            else:
                tag_word.append('not')
            lex_9 = morph.parse(word)[0].tag.tense  # время (настоящее, прошедшее, будущее)
            if lex_9 != None:
                tag_word.append(lex_9)
            else:
                tag_word.append('not')
            lex_10 = morph.parse(word)[0].tag.transitivity  # переходность (переходный, непереходный)
            if lex_10 != None:
                tag_word.append(lex_10)
            else:
                tag_word.append('not')
            lex_11 = morph.parse(word)[0].tag.voice # залог (действительный, страдательный)
            if lex_11 != None:
                tag_word.append(lex_11)
            else:
                tag_word.append('not')
            tags.append(tag_word)
            # генератором списков проходимся по каждому сформированному слову и лемматиризируем его с помощью pymorphy
          words = [morph.parse(word)[0].normal_form for word in words]

          return words, tags

        # Формирование обучающей выборки по листу индексов слов (разделение на короткие векторы)
        def getSetFromIndexes(wordIndexes, xLen, step):
          xTrain = []
          wordsLen = len(wordIndexes)
          index = 0
          while (index + xLen <= wordsLen):
            xTrain.append(wordIndexes[index:index+xLen])
            index += step
            
          return xTrain

        # Функция формирование обучающей и проверочной выборки выборки из двух листов индексов от двух классов
        def createSetsMultiClasses(wordIndexes, xLen, step):

          #Для каждого из 10 классов создаём обучающую выборку из индексов
          nClasses = len(wordIndexes)
          classesXTrain = []
          for wI in wordIndexes:
            classesXTrain.append(getSetFromIndexes(wI, xLen, step))

          #Формируем один общий xTrain
          xTrain = []
          yTrain = []
          
          for t in range(nClasses):
            xT = classesXTrain[t]
            for i in range(len(xT)):
              xTrain.append(xT[i])
            
            #Формируем yTrain по номеру класса
            currY = utils.to_categorical(t, nClasses)
            for i in range(len(xT)):
              yTrain.append(currY)

          xTrain = np.array(xTrain)
          yTrain = np.array(yTrain)
          
          return (xTrain, yTrain)

        # Загружаем текст 
        test_text = text_input

        # Здесь будут лежать все списки слов каждого из описаний заболеваний и списки списков граммем для каждого слова
        test_words = []
        test_tags = []
        words, tags = text2Words(test_text[-531:])
        test_words.append(words)
        test_tags.append(tags)

        # загружаем токенайзер слов
        with open('tokenizer1.json') as f:
            data = json.load(f)
            tokenizer = tokenizer_from_json(data)
        # загружаем токенайзер граммем
        with open('tokenizer2.json') as f:
            data = json.load(f)
            tokenizer2 = tokenizer_from_json(data)

        # Преобразовываем слова и тэги в последовательность индексов
        # Делаем короткие описания симптомов  длины xLen, оставляем
      

        #Тестовые тексты в индексы
        xTestIndexes = tokenizer.texts_to_sequences(test_words)

        if len(xTestIndexes[0]) < 50:
          
          xTestIndexes = pad_sequences(xTestIndexes, maxlen=xLen , padding='post', value=0)
          #Преобразовываем тэги в последовательность индексов согласно частотному словарю
          xTestTagsIndexes = []
          for tag in test_tags:  # так как теги имеют дополнительные вложенные списки поэтому итерируем в нужном для токенайзера формате
            xTestTagsIndexes.append(tokenizer2.texts_to_sequences(tag))
          xTestTagsIndexes = pad_sequences(xTestTagsIndexes, maxlen=xLen , padding='post', value=0)
        else:
          xTestTagsIndexes = []
          for tag in test_tags:  # так как теги имеют дополнительные вложенные списки поэтому итерируем в нужном для токенайзера формате
            xTestTagsIndexes.append(tokenizer2.texts_to_sequences(tag))

        # Создаем "раскусанные" выборки длины xLen для тестовой
        xTest, _ = createSetsMultiClasses(xTestIndexes, xLen, step)
        xTestTags, _ = createSetsMultiClasses(xTestTagsIndexes, xLen, step) 

        # Преобразовываем полученные тестовые выборки слов и тэгов из последовательности индексов в матрицы нулей и единиц по принципу Bag of Words
        xTest01 = tokenizer.sequences_to_matrix(xTest.tolist(), mode="tfidf") # Подаем xTest в виде списка чтобы метод успешно сработал
        xTestTags = np.reshape(xTestTags, (xTestTags.shape[0], -1)) # Делаем решейп для токенайзера
        xTestTags01 = tokenizer2.sequences_to_matrix(xTestTags.tolist())  

        # Загружаем модель с весами
        model = нейронка

        # Вывод предсказания в виде номера документа, списка заболеваний и вероятностей того, на сколько описание соответствует тому или иному заболеванию
        sumrec = [x * 0 for x in range(nClasses)]   #создаем список суммы значений предсказанных "окон" зополненный нулями по количеству классов обучающего датасета (в нашем случае их 10)
        meanrec = []                                #создаем пустой список средних значений предсказанных "окон"
        currPred = model.predict([xTest, xTest01, xTestTags01]) #предсказываем документ
        currOut = np.argmax(currPred, axis=1)       #определяем номер распознанного класса для каждого блока слов длины xLen
          
        # Проходим по каждому элементу предсказания блоков слов длины xLen и суммируем значения по каждому предсказанному классу в лист sumrec
        for i in range(len(currPred)):
          for j in range(len(currPred[i])):
            sumrec[j] += currPred[i][j]

        # Проходим по каждому элементу листа sumrec и записываем среднее значение вероятности предсказанных блоков слов длины xLen в лист meanrec по каждому классу
        for t in range(len(sumrec)):
          meanrec.append(round(sumrec[t]/len(currPred),4))
          
        # Если нужно то выводим вероятности по всем классам для данного текста
        for m in range(nClasses):
          print(classes[m], ' '* (12 - len(classes[m])),' - ', round(meanrec[m]*100, 2), '%')
          # if round(meanrec[m]*100, 2) > 30:
          #   print(classes[m], ' - ', round(meanrec[m]*100, 2), '%')
        print()
          

        #выводим на печать диагноз
        diagnos =  'Нейросеть думает, что это: ' + str(classes[int(sum(currOut)/len(currPred))]) + ' - ' + str(round(meanrec[int(sum(currOut)/len(currPred))]*100, 2)) + '%'
        print(diagnos)

def тест_модели_писатели(нейронка, xLen, step, симптомы, классы):
  predict_txt(нейронка, xLen, step, симптомы, классы)
  
def predict_txt(нейронка, xLen, step, text_input = None, классы=None):
    if text_input is None:
        diagnos =  "Текстовое поле пустое"
    
    else:       

        classes = классы
        nClasses = len(классы)
        stop_words = nltk.corpus.stopwords.words('russian')
        lexeme_list = ['POS', 'animacy', 'aspect', 'case', 'gender', 'involvement', 'mood', 'number', 'person', 'tense', 'transitivity', 'voice']

        # Функция очистки текста и превращение в набор слов
        def text2Words(text):
          text = text.replace(".", " ")
          text = text.replace("—", " ")
          text = text.replace(",", " ")
          text = text.replace("!", " ")
          text = text.replace("?", " ")
          text = text.replace("…", " ")
          text = text.replace("-", " ")
          text = text.replace("(", " ")
          text = text.replace(")", " ")
          text = text.replace(";", " ")
          text = text.replace("°c", " ")
          text = text.replace("–", " ")
          text = text.replace("и/ить", " ")
          text = text.lower()
          morph = pymorphy2.MorphAnalyzer()
          
          words = []
          tags = []
          currWord = ""
          for symbol in text[1:]:

            if (symbol != " "):
              currWord += symbol
            else:
              if (currWord != "") & (currWord not in stop_words): # 
                words.append(currWord)
                currWord = ""
              else:
                currWord = ""
          if (currWord != "") & (currWord not in stop_words): #
                words.append(currWord)
          # генератором списков проходимся по каждому сформированному слову и лемматиризируем его с помощью pymorphy до получения граммем
          for word in words:
            tag_word = []
            lex_0 = morph.parse(word)[0].tag.POS  # Part of Speech, часть речи
            if lex_0 != None:
                tag_word.append(lex_0)
            else:
                tag_word.append('not')
            lex_1 = morph.parse(word)[0].tag.animacy  # одушевленность
            if lex_1 != None:
                tag_word.append(lex_1)
            else:
                tag_word.append('not')
            lex_2 = morph.parse(word)[0].tag.aspect # вид: совершенный или несовершенный
            if lex_2 != None:
                tag_word.append(lex_2)
            else:
                tag_word.append('not')
            lex_3 = morph.parse(word)[0].tag.case # падеж
            if lex_3 != None:
                tag_word.append(lex_3)  
            else:
                tag_word.append('not')     
            lex_4 = morph.parse(word)[0].tag.gender # род (мужской, женский, средний)
            if lex_4 != None:
                tag_word.append(lex_4)
            else:
                tag_word.append('not') 
            lex_5 = morph.parse(word)[0].tag.involvement  # включенность говорящего в действие
            if lex_5 != None:
                tag_word.append(lex_5)
            else:
                tag_word.append('not')
            lex_6 = morph.parse(word)[0].tag.mood # наклонение (повелительное, изъявительное)
            if lex_6 != None:
                tag_word.append(lex_6)
            else:
                tag_word.append('not')
            lex_7 = morph.parse(word)[0].tag.number # число (единственное, множественное)
            if lex_7 != None:
                tag_word.append(lex_7)
            else:
                tag_word.append('not')
            lex_8 = morph.parse(word)[0].tag.person # лицо (1, 2, 3)
            if lex_8 != None:
                tag_word.append(lex_8)
            else:
                tag_word.append('not')
            lex_9 = morph.parse(word)[0].tag.tense  # время (настоящее, прошедшее, будущее)
            if lex_9 != None:
                tag_word.append(lex_9)
            else:
                tag_word.append('not')
            lex_10 = morph.parse(word)[0].tag.transitivity  # переходность (переходный, непереходный)
            if lex_10 != None:
                tag_word.append(lex_10)
            else:
                tag_word.append('not')
            lex_11 = morph.parse(word)[0].tag.voice # залог (действительный, страдательный)
            if lex_11 != None:
                tag_word.append(lex_11)
            else:
                tag_word.append('not')
            tags.append(tag_word)
            # генератором списков проходимся по каждому сформированному слову и лемматиризируем его с помощью pymorphy
          words = [morph.parse(word)[0].normal_form for word in words]

          return words, tags

        # Формирование обучающей выборки по листу индексов слов (разделение на короткие векторы)
        def getSetFromIndexes(wordIndexes, xLen, step):
          xTrain = []
          wordsLen = len(wordIndexes)
          index = 0
          while (index + xLen <= wordsLen):
            xTrain.append(wordIndexes[index:index+xLen])
            index += step
            
          return xTrain

        # Функция формирование обучающей и проверочной выборки выборки из двух листов индексов от двух классов
        def createSetsMultiClasses(wordIndexes, xLen, step):

          #Для каждого из 10 классов создаём обучающую выборку из индексов
          nClasses = len(wordIndexes)
          classesXTrain = []
          for wI in wordIndexes:
            classesXTrain.append(getSetFromIndexes(wI, xLen, step))

          #Формируем один общий xTrain
          xTrain = []
          yTrain = []
          
          for t in range(nClasses):
            xT = classesXTrain[t]
            for i in range(len(xT)):
              xTrain.append(xT[i])
            
            #Формируем yTrain по номеру класса
            currY = utils.to_categorical(t, nClasses)
            for i in range(len(xT)):
              yTrain.append(currY)

          xTrain = np.array(xTrain)
          yTrain = np.array(yTrain)
          
          return (xTrain, yTrain)

        # Загружаем текст 
        test_text = text_input

        # Здесь будут лежать все списки слов каждого из описаний заболеваний и списки списков граммем для каждого слова
        test_words = []
        test_tags = []
        words, tags = text2Words(test_text[-531:])
        test_words.append(words)
        test_tags.append(tags)

        # загружаем токенайзер слов
        with open('tokenizer1.json') as f:
            data = json.load(f)
            tokenizer = tokenizer_from_json(data)
        # загружаем токенайзер граммем
        with open('tokenizer2.json') as f:
            data = json.load(f)
            tokenizer2 = tokenizer_from_json(data)

        # Преобразовываем слова и тэги в последовательность индексов
        # Делаем короткие описания симптомов  длины xLen, оставляем
        #xLen = 50
        #step = 1

        #Тестовые тексты в индексы
        xTestIndexes = tokenizer.texts_to_sequences(test_words)

        if len(xTestIndexes[0]) < 50:
          
          xTestIndexes = pad_sequences(xTestIndexes, maxlen=xLen , padding='post', value=0)
          #Преобразовываем тэги в последовательность индексов согласно частотному словарю
          xTestTagsIndexes = []
          for tag in test_tags:  # так как теги имеют дополнительные вложенные списки поэтому итерируем в нужном для токенайзера формате
            xTestTagsIndexes.append(tokenizer2.texts_to_sequences(tag))
          xTestTagsIndexes = pad_sequences(xTestTagsIndexes, maxlen=xLen , padding='post', value=0)
        else:
          xTestTagsIndexes = []
          for tag in test_tags:  # так как теги имеют дополнительные вложенные списки поэтому итерируем в нужном для токенайзера формате
            xTestTagsIndexes.append(tokenizer2.texts_to_sequences(tag))

        # Создаем "раскусанные" выборки длины xLen для тестовой
        xTest, _ = createSetsMultiClasses(xTestIndexes, xLen, step)
        xTestTags, _ = createSetsMultiClasses(xTestTagsIndexes, xLen, step) 

        # Преобразовываем полученные тестовые выборки слов и тэгов из последовательности индексов в матрицы нулей и единиц по принципу Bag of Words
        xTest01 = tokenizer.sequences_to_matrix(xTest.tolist(), mode="tfidf") # Подаем xTest в виде списка чтобы метод успешно сработал
        xTestTags = np.reshape(xTestTags, (xTestTags.shape[0], -1)) # Делаем решейп для токенайзера
        xTestTags01 = tokenizer2.sequences_to_matrix(xTestTags.tolist())  

        # Загружаем модель с весами
        model = нейронка

        # Вывод предсказания в виде номера документа, списка заболеваний и вероятностей того, на сколько описание соответствует тому или иному заболеванию
        sumrec = [x * 0 for x in range(nClasses)]   #создаем список суммы значений предсказанных "окон" зополненный нулями по количеству классов обучающего датасета (в нашем случае их 10)
        meanrec = []                                #создаем пустой список средних значений предсказанных "окон"
        currPred = model.predict([xTest, xTest01, xTestTags01]) #предсказываем документ
        currOut = np.argmax(currPred, axis=1)       #определяем номер распознанного класса для каждого блока слов длины xLen
          
        # Проходим по каждому элементу предсказания блоков слов длины xLen и суммируем значения по каждому предсказанному классу в лист sumrec
        for i in range(len(currPred)):
          for j in range(len(currPred[i])):
            sumrec[j] += currPred[i][j]

        # Проходим по каждому элементу листа sumrec и записываем среднее значение вероятности предсказанных блоков слов длины xLen в лист meanrec по каждому классу
        for t in range(len(sumrec)):
          meanrec.append(round(sumrec[t]/len(currPred),4))
          
        # Если нужно то выводим вероятности по всем классам для данного текста
        for m in range(nClasses):
          print(classes[m], ' '* (12 - len(classes[m])),' - ', round(meanrec[m]*100, 2), '%')
          # if round(meanrec[m]*100, 2) > 30:
          #   print(classes[m], ' - ', round(meanrec[m]*100, 2), '%')
        print()
          

        #выводим на печать диагноз
        diagnos =  'Нейросеть считает что это: ' + str(classes[int(sum(currOut)/len(currPred))]) + ' - ' + str(round(meanrec[int(sum(currOut)/len(currPred))]*100, 2)) + '%'
        print(diagnos)
        
def создать_выборки_чатбота(вопросы, ответы, количество_пар=10000):
  ######################
  # Разбираем вопросы-ответы с проставлением тегов ответам
  ######################
  # Собираем вопросы и ответы в списки
  questions = вопросы[:количество_пар] # здесь будет список вопросов
  answers = ответы[:количество_пар  ] # здесь будет список ответов  
  assert len(questions) == len(answers), 'Количество вопросов не совпадает с количеством ответов'


  # Сделаем теги-метки для начала и конца ответов
  n_answers = list()
  for i in range(len(answers)):
    n_answers.append( '<START> ' + answers[i] + ' <END>' )

  ######################
  # Подключаем керасовский токенизатор и собираем словарь индексов
  ######################
  tokenizer = Tokenizer(oov_token='unknown')
  tokenizer.fit_on_texts(questions + n_answers) # загружаем в токенизатор список вопросов-ответов для сборки словаря частотности
  vocabularyItems = list(tokenizer.word_index.items()) # список с cодержимым словаря
  vocabularySize = len(vocabularyItems)+1 # размер словаря
  
  ######################
  # Устанавливаем закодированные входные данные(вопросы)
  ######################
  tokenizedQuestions = tokenizer.texts_to_sequences(questions) # разбиваем текст вопросов на последовательности индексов
  maxLenQuestions = max([ len(x) for x in tokenizedQuestions]) # уточняем длину самого большого вопроса
  # Делаем последовательности одной длины, заполняя нулями более короткие вопросы
  paddedQuestions = pad_sequences(tokenizedQuestions, maxlen=maxLenQuestions, padding='post')

  # Предподготавливаем данные для входа в сеть
  encoderForInput = np.array(paddedQuestions) # переводим в numpy массив

  ######################
  # Устанавливаем раскодированные входные данные(ответы)
  ######################
  tokenizedAnswers = tokenizer.texts_to_sequences(n_answers) # разбиваем текст ответов на последовательности индексов
  maxLenAnswers = max([len(x) for x in tokenizedAnswers]) # уточняем длину самого большого ответа
  # Делаем последовательности одной длины, заполняя нулями более короткие ответы
  paddedAnswers = pad_sequences(tokenizedAnswers, maxlen=maxLenAnswers, padding='post')

  # Предподготавливаем данные для входа в сеть
  decoderForInput = np.array(paddedAnswers) # переводим в numpy массив

  ######################
  # Раскодированные выходные данные(ответы)
  ######################
  tokenizedAnswers = tokenizer.texts_to_sequences(n_answers) # разбиваем текст ответов на последовательности индексов
  for i in range(len(tokenizedAnswers)) : # для разбитых на последовательности ответов
    tokenizedAnswers[i] = tokenizedAnswers[i][1:] # избавляемся от тега <START>
  # Делаем последовательности одной длины, заполняя нулями более короткие ответы
  paddedAnswers = pad_sequences(tokenizedAnswers, maxlen=maxLenAnswers , padding='post')

  #oneHotAnswers = utils.to_categorical(paddedAnswers, vocabularySize) # переводим в one hot vector
  oneHotAnswers = paddedAnswers # переводим в one hot vector
  decoderForOutput = np.array(oneHotAnswers) # и сохраняем в виде массива numpy
  tokenizer_json1 = tokenizer.to_json()
  with io.open('tokenizer_cb.json', 'w', encoding='utf-8') as f:
    f.write(json.dumps(tokenizer_json1, ensure_ascii=False))
  return (encoderForInput, decoderForInput), decoderForOutput, vocabularySize

def тест_модели_чат_бот(модель, размер_словаря, энкодер, декодер):
  with open('tokenizer_cb.json') as f:
    data = json.load(f)
    tokenizer = tokenizer_from_json(data)
    
  def makeInferenceModels(): 
    encoderInputs = Input(shape=(None , )) # размеры на входе сетки (здесь будет encoderForInput)
    layers = энкодер.split()
    if '-' in layers[0]:
      буква, параметр = layers[0].split('-')
      x = Embedding(размер_словаря, int(параметр), mask_zero=True) (encoderInputs)      
    for i in range(1, len(layers)-1):
      layer = создать_слой(layers[i])
      assert layer!=0, 'Невозможно добавить указанный слой: '+layer
      x = создать_слой(layers[i]) (x)
    if '-' in layers[-1]:
      буква, параметр = layers[-1].split('-')
    encoderOutputs, state_h , state_c = LSTM(int(параметр), return_state=True)(x)
    encoderStates = [state_h, state_c]
  
    encoderModel = Model(encoderInputs, encoderStates) 
    
    
    decoderStateInput_h = Input(shape=(None ,)) # обозначим размерность для входного слоя с состоянием state_h
    decoderStateInput_c = Input(shape=(None ,)) # обозначим размерность для входного слоя с состоянием state_c
    decoderStatesInputs = [decoderStateInput_h, decoderStateInput_c] # возьмем оба inputs вместе и запишем в decoderStatesInputs
    
    decoderInputs = Input(shape=(None, )) # размеры на входе сетки (здесь будет decoderForInput)
    layers = декодер.split()
    if '-' in layers[0]:
      буква, параметр = layers[0].split('-')
      x = Embedding(размер_словаря, int(параметр), mask_zero=True) (decoderInputs) 
    for i in range(1, len(layers)-1):
      layer = создать_слой(layers[i])
      assert layer!=0, 'Невозможно добавить указанный слой: '+layer
      x = создать_слой(layers[i]) (x)
    if '-' in layers[-1]:
      буква, параметр = layers[-1].split('-')
    decoderLSTM = LSTM(int(параметр), return_state=True, return_sequences=True)
  
    decoderOutputs, state_h, state_c = decoderLSTM(x, initial_state=decoderStatesInputs)
    decoderStates = [state_h, state_c] # LSTM даст нам новые состояния
    decoderDense = Dense(размер_словаря, activation='softmax') 
    decoderOutputs = decoderDense(decoderOutputs) # и ответы, которые мы пропустим через полносвязный слой с софтмаксом
    decoderModel = Model([decoderInputs] + decoderStatesInputs, [decoderOutputs] + decoderStates)
    return encoderModel , decoderModel

  def strToTokens(sentence: str): # функция принимает строку на вход (предложение с вопросом)
    words = sentence.lower().split() # приводит предложение к нижнему регистру и разбирает на слова
    tokensList = list() # здесь будет последовательность токенов/индексов
    for word in words: # для каждого слова в предложении
      try:
        tokensList.append(tokenizer.word_index[word]) # определяем токенизатором индекс и добавляем в список
      except:
        print('Я не знаю такого слова:', word)
        tokensList.append(0)
    return pad_sequences([tokensList], maxlen=13 , padding='post')
    

  encModel , decModel = makeInferenceModels() # запускаем функцию для построения модели кодера и декодера
  print('Тест общения с ботом. (Для завершения наберите «Выход»)')
  while(True):
      # Получаем значения состояний, которые определит кодер в соответствии с заданным вопросом
      ques = input( 'Задайте вопрос : ' )
      if ques=='Выход':
	      break
      statesValues = encModel.predict(strToTokens(ques))
      # Создаём пустой массив размером (1, 1)
      emptyTargetSeq = np.zeros((1, 1))    
      emptyTargetSeq[0, 0] = tokenizer.word_index['start'] # положим в пустую последовательность начальное слово 'start' в виде индекса

      stopCondition = False # зададим условие, при срабатывании которого, прекратится генерация очередного слова
      decodedTranslation = '' # здесь будет собираться генерируемый ответ
      while not stopCondition : # пока не сработало стоп-условие
        # В модель декодера подадим пустую последовательность со словом 'start' и состояния предсказанные кодером по заданному вопросу.
        # декодер заменит слово 'start' предсказанным сгенерированным словом и обновит состояния
        decOutputs , h , c = decModel.predict([emptyTargetSeq] + statesValues)
        
        #argmax пробежит по вектору decOutputs'а[0,0,15104], найдет макс.значение, и вернёт нам номер индекса под которым оно лежит в массиве
        sampledWordIndex = np.argmax( decOutputs[0, 0, :]) # argmax возьмем от оси, в которой 15104 элементов. Получили индекс предсказанного слова.
        sampledWord = None # создаем переменную, в которую положим слово, преобразованное на естественный язык
        for word , index in tokenizer.word_index.items():
          if sampledWordIndex == index: # если индекс выбранного слова соответствует какому-то индексу из словаря
            decodedTranslation += ' {}'.format(word) # слово, идущее под этим индексом в словаре, добавляется в итоговый ответ 
            sampledWord = word # выбранное слово фиксируем в переменную sampledWord
        
        # Если выбранным словом оказывается 'end' либо если сгенерированный ответ превышает заданную максимальную длину ответа
        if sampledWord == 'end' or len(decodedTranslation.split()) > 13:
          stopCondition = True # то срабатывает стоп-условие и прекращаем генерацию

        emptyTargetSeq = np.zeros((1, 1)) # создаем пустой массив
        emptyTargetSeq[0, 0] = sampledWordIndex # заносим туда индекс выбранного слова
        statesValues = [h, c] # и состояния, обновленные декодером
        # и продолжаем цикл с обновленными параметрами
      
      print(decodedTranslation[:-3]) # выводим ответ сгенерированный декодером

def разметка_договора(model, корректор):
  xLen = 256
  step = 30
  embeddingSize = 300
  t = корректор

  test_file = '/content/Договоры/test_doc.txt' # Путь к папке с договором
  
  ################## ПОВТОР ВЫШЕ ИМЕЮЩИХСЯ ФУНКЦИЙ из лекции ###########################################
  # Чтение файла текста из файла, очитска от знаков препинания
  def readText(fileName):
    f = open(fileName, 'r') #Открываем наш файл для чтения и считываем из него данные 
    text = f.read() #Записываем прочитанный текст в переменную 
    # Определяем символы для удаления
    delSymbols = ['\n', "\t", "\ufeff", ".", "_", "-", ",", "!", "?", "–", "(", ")", "«", "»", "№", ";"]

    for dS in delSymbols: # Каждый символ в списке символов для удаления
      text = text.replace(dS, " ") # Удаляем, заменяя на пробел

    # Ищем шаблоны и символы в строке и меняем на указанную подстроку
    text = re.sub("[.]", " ", text) 
    text = re.sub(":", " ", text)
    text = re.sub("<", " <", text)
    text = re.sub(">", "> ", text)

    # Метод split разделит текст по пробелам (а их может быть несколько после удаления символов)
    # При помощи метода join запишем все разделенные слова снова в строку
    text = ' '.join(text.split()) 

    text = text.lower() # Переводим текст в нижний регистр
    return text # Возвращаем тексты

  # Преобразование исходного текста в список из слов в нормальной форме 
  def text2Words(text):
    morph = pymorphy2.MorphAnalyzer() # Создаем экземпляр класса MorphAnalyzer
    words = text.split(' ') # Разделяем текст на пробелы
    words = [morph.parse(word)[0].normal_form for word in words] #Переводим каждое слово в нормалную форму  
    return words # Возвращаем слова


  def get01XSamples(tok_agreem, tags_index):
    tags01 = [] # Список для тегов
    indexes = [] # Здесь будут лежать индексы
  
    for agreement in tok_agreem: # Проходимся по каждому договору-списку
      tag_place = [0, 0, 0, 0, 0, 0] # Создаем вектор [0,0,0,0,0,0]
      for ex in agreement: # Проходимся по каждому слову договора
          if ex in tags_index: # Смотрим, если индекс оказался нашим тегом
            place = np.argwhere(tags_index==ex) # Записываем под каким местом лежит этот тег в своем списке
            if len(place)!=0: # Проверяем, чтобы тег действительно был
              if place[0][0]<6: # Первые шесть тегов в списке - открывающие
                tag_place[place[0][0]] = 1    # Поэтому ставим 1
              else: 
                tag_place[place[0][0] - 6] = 0  # Остальные теги закрывающие, так что меняем на ноль
          else:          
            tags01.append(tag_place.copy()) # Расширяем наш список с каждой итерацией. Получаем в конце длинный список из всех тегов в одном 
            indexes.append(ex) # Докидываем индекс слова в список индексов

    return indexes, tags01

  # Получение списка слов из индексов
  def reverseIndex(clean_voc, x):
    reverse_word_map = dict(map(reversed, clean_voc.items())) # Берем пары значений всего словаря и размечаем наоборот, т.е. value:key
    words = [reverse_word_map.get(letter) for letter in x] # Вытаскиваем по каждому ключу в список
    return words # Возвращаем полученный текст


  # Формируем выборку из индексов
  def getSetFromIndexes(wordIndexes, xLen, step): 
    xBatch = [] # Лист для фрагментов текста
    wordsLen = len(wordIndexes) # Определяем длинну текста
    index = 0 # Задаем стартовый индекс
    
    while (index + xLen <= wordsLen): # Пока сумма индекса с длинной фрагмента меньше или равна числу слов в выборке
      xBatch.append(wordIndexes[index:index+xLen]) # Добавляем X в лист фразментов текста
      index += step # Сдвигаемся на step

    return xBatch # Лист для фрагментов текста

  # Создаем выборки
  def getSets(model, senI, tagI):
    xVector = [] # Здесь будет лежать embedding представление каждого из индексов
    tmp = [] # Временный список
    for text in senI: # Проходимся по каждому тексту-списку
      tmp=[]
      for word in text: # Проходимся по каждому слову в тексте-списке
        tmp.append(model[word]) 

      xVector.append(tmp)

    return np.array(xVector), np.array(tagI)
  ############################################################################################################

  # Получение договора для последующей разметки
  test_dogovor = [] # Список, в который запишем все наши договоры test_dogovor
  txt = readText(test_file) # Читаем текст договора
  if txt != '': # Если текст не пустой
    test_dogovor.append(readText(test_file)) # Преобразуем файл в одну строку и добавляем в agreements

  # покажем текст
  n = 0
  print(os.listdir(test_directory)[n], '\n', '\n') # Выведем название n-ого договора
  #test_dogovor[n] 

  #переводим в слова
  words_test_dogovor = [] # Здесь будут храниться все договора в виде списка слов
  curTime = time.time() # Засечем текущее время
  for i in range(len(test_dogovor)): # Проходимся по всем договорам
    words_test_dogovor.append(text2Words(test_dogovor[i])) # Преобразуем очередной договор в список слов и добавляем в words

  import textwrap
  print("Полученный и очищенный от лишних знаков договор", '\n')
  text_vnachale = ' '.join(words_test_dogovor[0])
  print(textwrap.fill(text_vnachale, 100), '\n', '\n')

  # Преобразовываем текст в последовательность индексов согласно частотному словарю
  test_dogovor_tok_agreem = tokenizer.texts_to_sequences(words_test_dogovor) # Обучающие тесты в индексы

  # decoded_text => test_dogovor_decoded
  test_dogovor_xData, test_dogovor_yData = get01XSamples(test_dogovor_tok_agreem,tags_index) # Распознаем теги и создаем список с ними, с индексами
  test_dogovor_decoded = reverseIndex(clean_voc, test_dogovor_xData) # Для создания списков с embedding-ами сначала преобразуем список индексов обратно в слова

  # Генерируем наборы с заданными параметрами окна
  test_dogovor_xTest = getSetFromIndexes(test_dogovor_decoded, xLen, step) # Последовательность из xLen слов
  test_dogovor_yTest = getSetFromIndexes(test_dogovor_yData, xLen, step) # Последовательность из xLen-тегов
  modelGENSIM = word2vec.Word2Vec(test_dogovor_xTest, size = embeddingSize, window = 10, min_count = 1, workers = 10, iter = 10)

  test_dogovor_GENSIM, y_test_dogovor_GENSIM = getSets(modelGENSIM, test_dogovor_xTest, test_dogovor_yTest)

  pred = model(test_dogovor_GENSIM) # получаем предикт от модели
  print('Категории по которым размечаются договоры', '\n')
  print('s1 - Условия')
  print('s2 - Запреты')
  print('s3 - Стоимость, всё про цены и деньги')
  print('s4 - Всё про сроки')
  print('s5 - Неустойка')
  print('s6 - Всё про адреса и геолокации', '\n', '\n')


  # функция реконструирует из предикта в лист векторов 0 и 1
  def reconst_predict(pred, t):
    pred_new = np.zeros_like(pred)
    mask = np.where(pred > t) # Записываем маску в при сравнении с корректоромзаданным в скобках функции
    pred_new[mask] = 1  # присваиваеи 1 где выне порога заданного корректором

    k = xLen//step                      # сколько целых шагов step в наборе xLen
    p = len(test_dogovor_yData)//xLen   # сколько целых шагов xLen в договоре test_dogovor_yData
    reconst = pred_new[0]               # берем 1й кусок xLen
    for i in range(1, p):              
      reconst = np.append(reconst, pred_new[i*k], axis = 0)   # берем через xLen еще и  еще  кусок xLen  p раз
    end = pred_new[-1][:(len(test_dogovor_yData)-p*xLen)]     # вычисляем сколько нужно в конце до длины test_dogovor_yData
    reconst_lst = np.append(reconst, end, axis = 0)           # докидываем в конце остаток

    return reconst_lst


  #списки тегов для функций ниже
  open_tags = ['<s' + str(i) + '>' for i in range(1, 7)] # Получаем список открывающих тегов
  all_tags = open_tags
  close_tags = ['</s' + str(i) + '>' for i in range(1, 7)] # Получаем список закрывающих тегов
  all_tags.extend(close_tags)

  # функция обратного преоброзавания списка из векторов 0 и1 собранного из предикта
  def get_list_with_tags(reconst_zone):
    # вставка тегов
    super_lst = []
    check = [0,0,0,0,0,0]
    for i in range(len(reconst_zone)):
      for j in range(len(reconst_zone[i])):
        if (reconst_zone[i][j] == 1 and check[j] == 0):
          super_lst.append(open_tags[j])
          check[j] = 1
        if (reconst_zone[i][j] == 1 and check[j] == 1):
          pass
        if (reconst_zone[i][j] == 0 and check[j] == 1):
          super_lst.append(close_tags[j])
          check[j] = 0
        if (reconst_zone[i][j] == 0 and check[j] == 0):
          pass
      super_lst.append(0)
    list_with_tags = super_lst   

    return list_with_tags


  # функция разметки ОДНОГО ПЕРВОГО договора тегами, под демонстрацию!!!!
  def razmetka(dogovor_lst, list_with_tags):
    dogovor = dogovor_lst[0]
    out_list=[]
    s = 0
    k = len(dogovor)/len(list_with_tags) # dogovor и list_with_tags - эти массивы обычно не совпадают
    # и нам нужно перенести разбиение на зоны в одном в  разбиение на зоны в другом
    for j in range(len(list_with_tags)):
      i = round(j*k)                    # подгонка одного к другоиу
      for tags in all_tags:
        if i < len(list_with_tags):     # проверка на длину и текущий индекс
          if list_with_tags[i] == tags:
            out_list.extend(dogovor[s:i])
            out_list.append(list_with_tags[i])
            s = i
    if s < len(dogovor): # проверка на длину и финальный индекс
      out_list.extend(dogovor[s:])      # пристыковка финальной части если осталась
          
    return out_list

  recont_list = reconst_predict(pred, t)

  list_with_tags = get_list_with_tags(recont_list)

  dogovor_s_razmetka = razmetka(words_test_dogovor,list_with_tags)

  text_razmet = ' '.join(dogovor_s_razmetka)

  print("Размеченый нейронкой договор", '\n')
  print(textwrap.fill(text_razmet, 100), '\n', '\n')


  #Получение ранее размеченного вручную договора для сравнения
  test_dogovor_1 = [] # Список, в который запишем все наши договоры test_dogovor
  for filename in os.listdir(test_directory_1): # Проходим по всем файлам в директории договоров
    txt = readText(test_directory_1 + filename) # Читаем текст договора
    if txt != '': # Если текст не пустой
      test_dogovor_1.append(readText(test_directory_1 + filename)) # Преобразуем файл в одну строку и добавляем в agreements

  #переводим в слова
  words_test_dogovor_1 = [] # Здесь будут храниться все договора в виде списка слов
  curTime = time.time() # Засечем текущее время
  for i in range(len(test_dogovor_1)): # Проходимся по всем договорам
    words_test_dogovor_1.append(text2Words(test_dogovor_1[i])) # Преобразуем очередной договор в список слов и добавляем в words

  print("Размеченный ранее вручную договор для сравнения", '\n')
  text_vnachale = ' '.join(words_test_dogovor_1[0])
  print(textwrap.fill(text_vnachale, 100))


def загрузить_предобученную_модель():
  url = 'https://storage.googleapis.com/aiu_bucket/tokenizer_best.json'
  output = 'tokenizer_best.json' # Указываем имя файла, в который сохраняем файл
  gdown.download(url, output, quiet=True) # Скачиваем файл по указанному URL

  with open('tokenizer_best.json') as f:
    data = json.load(f)
    токенайзер = tokenizer_from_json(data)
  # url = ''
  # output = 'model_chatbot_100epochs(rms)+50(ada).h5'
  # gdown.download(url, output, quiet=True)  
  url = 'https://storage.googleapis.com/aiu_bucket/model_chatbot_100epochs(rms)%2B50(ada).h5'
  output = 'model_chatbot_100epochs(rms)%2B50(ada).h5' # Указываем имя файла, в который сохраняем файл
  gdown.download(url, output, quiet=True) # Скачиваем файл по указанному URL
  model = load_model('model_chatbot_100epochs(rms)%2B50(ada).h5')


  def strToTokens(sentence: str): # функция принимает строку на вход (предложение с вопросом)
    words = sentence.lower().split() # приводит предложение к нижнему регистру и разбирает на слова
    tokensList = list() # здесь будет последовательность токенов/индексов
    for word in words: # для каждого слова в предложении
      try:
        tokensList.append(токенайзер.word_index[word]) # определяем токенизатором индекс и добавляем в список
      except KeyError:
        pass
    # Функция вернёт вопрос в виде последовательности индексов, ограниченной длиной самого длинного вопроса из нашей базы вопросов
    return pad_sequences([tokensList], maxlen=13, padding='post')
  ######################
  # Устанавливаем связи между слоями рабочей модели и предобученной
  ######################
  def loadInferenceModels():
    encoderInputs = model.input[0]   # входом энкодера рабочей модели будет первый инпут предобученной модели(input_1)
    encoderEmbedding = model.layers[2] # связываем эмбединг слои(model.layers[2] это embedding_1)
    encoderOutputs, state_h_enc, state_c_enc = model.layers[4].output # вытягиваем аутпуты из первого LSTM слоя обуч.модели и даем энкодеру(lstm_1)
    encoderStates = [state_h_enc, state_c_enc] # ложим забранные состояния в состояния энкодера
    encoderModel = Model(encoderInputs, encoderStates) # формируем модель

    decoderInputs = model.input[1]   # входом декодера рабочей модели будет второй инпут предобученной модели(input_2)
    decoderStateInput_h = Input(shape=(200 ,)) # обозначим размерность для входного слоя с состоянием state_h
    decoderStateInput_c = Input(shape=(200 ,)) # обозначим размерность для входного слоя с состоянием state_c

    decoderStatesInputs = [decoderStateInput_h, decoderStateInput_c] # возьмем оба inputs вместе и запишем в decoderStatesInputs

    decoderEmbedding = model.layers[3] # связываем эмбединг слои(model.layers[3] это embedding_2)
    decoderLSTM = model.layers[5] # связываем LSTM слои(model.layers[5] это lstm_2)
    decoderOutputs, state_h, state_c = decoderLSTM(decoderEmbedding.output, initial_state=decoderStatesInputs)
    decoderStates = [state_h, state_c] # LSTM даст нам новые состояния

    decoderDense = model.layers[6] # связываем полносвязные слои(model.layers[6] это dense_1)
    decoderOutputs = decoderDense(decoderOutputs) # выход с LSTM мы пропустим через полносвязный слой с софтмаксом

      # Определим модель декодера, на входе далее будут раскодированные ответы (decoderForInputs) и состояния
      # на выходе предсказываемый ответ и новые состояния
    decoderModel = Model([decoderInputs] + decoderStatesInputs, [decoderOutputs] + decoderStates)
    return encoderModel , decoderModel

  ######################
  # Устанавливаем окончательные настройки и запускаем рабочую модель над предобученной
  ######################

  encModel , decModel = loadInferenceModels() # запускаем функцию для построения модели кодера и декодера
  
  display.clear_output(wait=True)
  print()
  print('Тест общения с ботом. (Для завершения наберите «Выход»)')
  while(True):
    # Получаем значения состояний, которые определит кодер в соответствии с заданным вопросом
    ques = input( 'Задайте вопрос : ' )
    if ques=='Выход':
	    break
    statesValues = encModel.predict(strToTokens(ques))
    # Создаём пустой массив размером (1, 1)
    emptyTargetSeq = np.zeros((1, 1))    
    emptyTargetSeq[0, 0] = токенайзер.word_index['start'] # положим в пустую последовательность начальное слово 'start' в виде индекса

    stopCondition = False # зададим условие, при срабатывании которого, прекратится генерация очередного слова
    decodedTranslation = '' # здесь будет собираться генерируемый ответ
    while not stopCondition : # пока не сработало стоп-условие
      # В модель декодера подадим пустую последовательность со словом 'start' и состояния предсказанные кодером по заданному вопросу.
      # декодер заменит слово 'start' предсказанным сгенерированным словом и обновит состояния
      decOutputs , h , c = decModel.predict([emptyTargetSeq] + statesValues)
      
      #argmax пробежит по вектору decOutputs'а[0,0,15104], найдет макс.значение, и вернёт нам номер индекса под которым оно лежит в массиве
      sampledWordIndex = np.argmax( decOutputs[0, 0, :]) # argmax возьмем от оси, в которой 15104 элементов. Получили индекс предсказанного слова.
      sampledWord = None # создаем переменную, в которую положим слово, преобразованное на естественный язык
      for word , index in токенайзер.word_index.items():
        if sampledWordIndex == index: # если индекс выбранного слова соответствует какому-то индексу из словаря
          decodedTranslation += ' {}'.format(word) # слово, идущее под этим индексом в словаре, добавляется в итоговый ответ 
          sampledWord = word # выбранное слово фиксируем в переменную sampledWord
      
      # Если выбранным словом оказывается 'end' либо если сгенерированный ответ превышает заданную максимальную длину ответа
      if sampledWord == 'end' or len(decodedTranslation.split()) > 13:
        stopCondition = True # то срабатывает стоп-условие и прекращаем генерацию

      emptyTargetSeq = np.zeros((1, 1)) # создаем пустой массив
      emptyTargetSeq[0, 0] = sampledWordIndex # заносим туда индекс выбранного слова
      statesValues = [h, c] # и состояния, обновленные декодером
      # и продолжаем цикл с обновленными параметрами
    
    print(decodedTranslation[:-3]) # выводим ответ сгенерированный декодером


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle 
import seaborn as sns
import utils
sns.set_style('darkgrid')

def clear_data(data):
    del data['Date']
    data = data.dropna() # удаляем строки с NaN
    return data

def getData():
  train_dataset = pd.read_csv('train_dataset.csv', sep=",")
  val_dataset = pd.read_csv('val_dataset.csv', sep=",")  
  train_dataset = clear_data(train_dataset)
  val_dataset = clear_data(val_dataset)
  return train_dataset, val_dataset

def show_data(data, start = 0, end = 0, param=['Close']):
    if end==0:
      end = data.shape[0]
    fig = plt.figure(figsize=(18,9))
    ax = fig.add_subplot(111) 
    for i in param:
      ax.plot(data[i][start:end], label=i)
    plt.legend()  
    plt.show() 

def show_full_data(train, val, param=['Close']):  
    data = train.append(val).reset_index() 
    fig = plt.figure(figsize=(18,9))
    ax = fig.add_subplot(111)
    for i in param:
      ax.plot(data[i], label=i)
    ax.add_patch( Rectangle((0, 600), 
                          train.shape[0], -600, 
                          fc ='none',  
                          ec ='g', 
                          lw = 2) )
    ax.add_patch( Rectangle((train.shape[0]+4, 600), 
                          val.shape[0]-4, -600, 
                          fc ='none',  
                          ec ='y', 
                          lw = 2) )
    ax.add_patch( Rectangle((train.shape[0]+val.shape[0]+8, 600), 
                          296, -600, 
                          fc ='none',  
                          ec ='b', 
                          lw = 2) )
    plt.legend()  
    plt.show() 